import { CalculatorResults } from "./calculator-results";
import { CalculatedParamResults } from "./param-calc-results";
import { NgParameter } from "../formulaire/ngparam";
import { ResultElement } from "jalhyd";
import { ServiceFactory } from "../services/service-factory";
import { PlottableData } from "./plottable-data";
import { GraphType } from "./graph-type";

export class VarResults extends CalculatedParamResults implements PlottableData {
    /**
     * paramètres variés
     */
    private _variatedParams: NgParameter[];

    /**
     * titre des colonnes des résultats variés
     */
    private _variableParamHeaders: string[];

    /**
     * clés des résultats complémentaires
     */
    public extraResultKeys: string[];

    /**
     * entête des colonnes des résultats supplémentaires
     */
    private _extraResultHeaders: string[];

    /**
     * type de graphe
     */
    public graphType: GraphType = GraphType.Scatter;

    /**
     * variated parameter or result displayed as chart's X-axis
     */
    public chartX: string;
    /**
     * variated parameter or result displayed as chart's Y-axis
     */
    public chartY: string;

    /** size of the longest variated parameter */
    public size: number;

    /** index of the longest variated parameter */
    public longest: number;

    /**
     * tableau des ordonnées du graphe des résultats variés
     */
    private _yValues: number[] = [];

    constructor() {
        super();
        this.reset();
    }

    public reset() {
        super.reset();
        this._variableParamHeaders = [];
        this._extraResultHeaders = [];
        this.extraResultKeys = [];
        this._yValues = [];
        this.longest = 0;
    }

    public get variatedParameters(): NgParameter[] {
        return this._variatedParams;
    }

    public set variatedParameters(p: NgParameter[]) {
        this._variatedParams = p;
        this._variableParamHeaders = this._variatedParams.map((v) => {
            return CalculatorResults.paramLabel(v, true);
        });
    }

    public get variableParamHeaders() {
        return this._variableParamHeaders;
    }

    public get yValues() {
        return this._yValues;
    }

    public get resultElements(): ResultElement[] {
        return this.result.resultElements;
    }

    public get extraResultHeaders() {
        return this._extraResultHeaders;
    }

    public getChartAxisLabel(symbol: string): string {
        // 1. calculated param ?
        if (this.calculatedParameter && this.calculatedParameter.symbol === symbol) {
            return this.calculatedParameterHeader;
        }
        // 2. variated param ?
        for (let i = 0; i < this.variatedParameters.length; i++) {
            if (this._variatedParams[i].symbol === symbol) {
                return this.variableParamHeaders[i];
            }
        }
        // 3. Result element
        return this.expandLabelFromSymbol(symbol);

    }

    /**
     * Returns the translated name of the given symbol (usually an extraResult) with
     * its unit, but without the symbol itself
     */
    public expandLabelFromSymbol(symbol: string): string {
        // calculator type for translation
        const sn = this.result.sourceNub;
        let ct = sn.calcType;
        if (sn.parent) {
            ct = sn.parent.calcType;
        }
        return ServiceFactory.instance.formulaireService.expandVariableNameAndUnit(ct, symbol);
    }

    /**
     * Returns the series of values for the required variated parameter / result element
     * @param symbol parameter / result symbol (ex: "Q")
     */
    public getValuesSeries(symbol: string) {
        const series = [];
        // 1. calculated param ?
        if (this._calculatedParam && this._calculatedParam.symbol === symbol) {
            if (this.result) {
                for (const r of this.result.resultElements) {
                    series.push(r.vCalc);
                }
            }
        }
        // 2. variated param ?
        for (let i = 0; i < this.variatedParameters.length; i++) {
            if (this._variatedParams[i].symbol === symbol) {
                const iter = this.variatedParameters[i].getExtendedValuesIterator(this.size);
                for (const v of iter) {
                    series.push(v);
                }
            }
        }
        // 3. Result element ?
        for (const r of this.result.resultElements) { // re:ResultElement
            for (const k in r.extraResults) {
                if (k === symbol) {
                    series.push(r.extraResults[k]);
                }
            }
        }

        return series;
    }

    /**
     * Returns a list of plottable parameters / result elements, that can be defined
     * as X or Y chart axis
     */
    public getAvailableChartAxis(): string[] {
        const res: string[] = [];
        if (this.calculatedParameter) {
            res.push(this.calculatedParameter.symbol);
        }
        for (const v of this._variatedParams) {
            res.push(v.symbol);
        }
        for (const erk of this.extraResultKeys) {
            res.push(erk);
        }
        return res;
    }

    /**
     * Returns the list of variating parameters
     * (used by tooltip functions)
     */
    public getVariatingParametersSymbols(): string[] {
        return this._variatedParams.map((vp) => {
            return vp.symbol;
        });
    }

    public update(displaySymbol: boolean) {
        if (this._variableParamHeaders.length === 0) {
            this._variableParamHeaders = this._variatedParams.map((v) => {
                return CalculatorResults.paramLabel(v, true);
            });
        }

        // liste la plus longue
        this.size = 0;
        let i = 0;
        for (const v of this._variatedParams) {
            const s = v.valuesIterator.count();
            if (s > this.size) {
                this.size = s;
                this.longest = i;
            }
            i++;
        }

        // valeurs du paramètre à calculer
        for (const r of this.result.resultElements) {
            this._yValues.push(r.vCalc);
        }

        // clés des résultats supplémentaires
        if (this.extraResultKeys.length === 0) {
            for (const re of this.result.resultElements) { // re:ResultElement
                for (const erk in re.extraResults) {
                    if (!this.extraResultKeys.includes(erk)) {
                        this.extraResultKeys.push(erk);
                    }
                }
            }
        }

        // set axis selectors values the first time
        let defaultY = this.chartY;
        if (this.calculatedParameter) {
            defaultY = this.calculatedParameter.symbol;
        } else if (this.extraResultKeys.length > 0) {
            defaultY = this.extraResultKeys[0];
        }
        this.chartX = this.chartX || this.variatedParameters[this.longest].symbol;
        this.chartY = defaultY;

        // calculator type for translation
        const sn = this.result.sourceNub;
        let ct = sn.calcType;
        if (sn.parent) {
            ct = sn.parent.calcType;
        }
        // entêtes des résultats supplémentaires
        this._extraResultHeaders = [];
        for (const k of this.extraResultKeys) {
            this._extraResultHeaders.push(
                ServiceFactory.instance.formulaireService.expandVariableNameAndUnit(ct, k)
            );
        }

        // when variable parameter changes, ensure the X / Y current values are still available
        // (might be the previous variated parameter, that is not accessible anymore)
        const aca = this.getAvailableChartAxis();
        if (! aca.includes(this.chartX)) {
            this.chartX = this.variatedParameters[0].symbol;
        }
        if (! aca.includes(this.chartY)) {
            this.chartY = this.variatedParameters[0].symbol;
        }
    }
}
